﻿using Antlr4.Runtime;
using Antlr4.Runtime.Tree;
using Antlr4.Runtime.Misc;
using Eingabeverwaltung.Metamodell;
using Eingabeverwaltung.Parser.Task_Parser;
using System.Collections.Generic;

namespace Eingabeverwaltung.Parser
{
    public class Task_Listener : ITaskListener
	{
        // object on which will be build
        private Container container;

        // helping Listener for building
        private Listener_Helper LH;

        // Task parts
        private List<TaskLemma> lemmata;
        private SingleTask mainTask;
        private List<SingleTask> lemmaTasks;
        private Variable indVar;

        // config parts
        private IConfig mainConfig;
        private SingleConfig singleConfig;
        private List<InductionCaseConfig> indCaseConfigs;
        private SingleConfig indHypConfig;
        private int maxPt;
        private int minSteps;
        private int maxSteps;

        /// <summary>
        /// Requires a Container with filled ADT
        /// </summary>
        /// <param name="container"></param>
        public Task_Listener(Container container, Listener_Helper LH)
		{
            this.container = container;

            this.LH = LH;

            lemmata = new List<TaskLemma>();
            mainTask = null;
            lemmaTasks = new List<SingleTask>();
            indVar = null;

            mainConfig = null;
            singleConfig = null;
            indCaseConfigs = null;
            indHypConfig = null;
            maxPt = -1;
            minSteps = -1;
            maxSteps = -1;
		}

        public void EnterTask([NotNull] TaskParser.TaskContext context) { }
        public void ExitTask([NotNull] TaskParser.TaskContext context) 
        {
            container.Task = new Task(mainTask,lemmaTasks,lemmata);
        }
        public void EnterTask_eq([NotNull] TaskParser.Task_eqContext context) { }
        public void ExitTask_eq([NotNull] TaskParser.Task_eqContext context)
        {
            if (mainConfig == null) LH.errors.Add(new SemanticAnalysisError(new Coords(context),"No Config found for main task!"));
            mainTask = new SingleTask("main task", LH.equation, LH.fixedVars, mainConfig, indVar);
            indVar = null;
        }

        public void EnterLemma_tasks([NotNull] TaskParser.Lemma_tasksContext context) { }
        public void ExitLemma_tasks([NotNull] TaskParser.Lemma_tasksContext context) { }
        public void EnterLemma_task([NotNull] TaskParser.Lemma_taskContext context) { }
        public void ExitLemma_task([NotNull] TaskParser.Lemma_taskContext context)
        {
            Coords coords = new Coords(context);
            string lemmaName = context.BEZ().GetText();
            Test_Lemmata_all_Unique(lemmaName,coords); 
            if (mainConfig == null) LH.errors.Add(new SemanticAnalysisError(coords,"No Config found for lemma " + lemmaName));
            lemmaTasks.Add(new SingleTask(lemmaName,LH.equation, LH.fixedVars, mainConfig, indVar));
            indVar = null;
        }

        public void EnterTask_lemmata([NotNull] TaskParser.Task_lemmataContext context) { }
        public void ExitTask_lemmata([NotNull] TaskParser.Task_lemmataContext context) { }
        public void EnterTask_lemma([NotNull] TaskParser.Task_lemmaContext context) { }
        public void ExitTask_lemma([NotNull] TaskParser.Task_lemmaContext context) 
        {
            string lemmaName = context.BEZ().GetText();
            Test_Lemmata_all_Unique(lemmaName,new Coords(context));
            lemmata.Add(new TaskLemma(lemmaName,LH.equation, LH.fixedVars));
        }

        // Induction
        public void EnterInd_var([NotNull] TaskParser.Ind_varContext context) 
        {
            string variableName = context.BEZ().GetText();
            Variable variable = LH.findVariableInADT_Vars(variableName);
            Coords coords = new Coords(context);
            if (!LH.allQVars.Contains(variable))
                LH.errors.Add(new SemanticAnalysisError(coords,"Induction variable for task "+ context.Parent.GetChild(0).GetText()+" was not definied as allquantified before."));
            else 
                if (variable.sort.GetConstructors().Count == 0)
                    LH.errors.Add(new SemanticAnalysisError(coords,"Induction variable for task " + context.Parent.GetChild(0).GetText()+" has no constructors!"));
            indVar = variable;
        }
        public void ExitInd_var([NotNull] TaskParser.Ind_varContext context) { }

        // Config
        public void EnterConfig([NotNull] TaskParser.ConfigContext context) { mainConfig = null; }
        public void ExitConfig([NotNull] TaskParser.ConfigContext context) 
        {
            if (mainConfig == null && singleConfig != null) mainConfig = singleConfig;
        }
        public void EnterInd_pt([NotNull] TaskParser.Ind_ptContext context)
        {
            indCaseConfigs = new List<InductionCaseConfig>();
            indHypConfig = null;
        }

        public void ExitInd_pt([NotNull] TaskParser.Ind_ptContext context)
        {
            Coords coords = new Coords(context);
            string taskName = context.Parent.Parent.GetChild(0).GetText();
            if (indVar == null)
                LH.errors.Add(new SemanticAnalysisError(coords,"Config for induction can't find its relation to induction variable!"));
            else if (indCaseConfigs.Count < indVar.sort.GetConstructors().Count)
                    LH.errors.Add(new SemanticAnalysisError(coords,"There are " + (indVar.sort.GetConstructors().Count - indCaseConfigs.Count)
                        + " missing cases for Config of induction for task " + taskName));
            if (indHypConfig == null) LH.errors.Add(new SemanticAnalysisError(coords,"No Config for Inductionhypotheses found for task " + taskName));
            mainConfig = new InductionConfig(indCaseConfigs, indHypConfig);
        }
        public void EnterCase_pt([NotNull] TaskParser.Case_ptContext context) { singleConfig = null; }
        public void ExitCase_pt([NotNull] TaskParser.Case_ptContext context)
        {
            Coords coords = new Coords(context);
            string constructorName = context.BEZ().GetText();
            Operation constructor = LH.findOperationInADT_Ops(constructorName, 0, false,coords);
            if (constructor as Constructor == null) LH.errors.Add(new SemanticAnalysisError(coords,"The induction case config refers not to a constructor: " + constructorName + "!"));
            if (singleConfig == null) LH.errors.Add(new SemanticAnalysisError(coords,"No config found for induction case " + constructorName + "!"));
            indCaseConfigs.Add(new InductionCaseConfig(constructor as Constructor, singleConfig));
        }
        public void EnterIh_pt([NotNull] TaskParser.Ih_ptContext context) { maxPt = -1; minSteps = -1; maxSteps = -1; }
        public void ExitIh_pt([NotNull] TaskParser.Ih_ptContext context) { indHypConfig = new SingleConfig(maxPt, minSteps, maxSteps); }

        public void EnterTask_pt([NotNull] TaskParser.Task_ptContext context) { maxPt = -1; minSteps = -1; maxSteps = -1; }
        public void ExitTask_pt([NotNull] TaskParser.Task_ptContext context)
        {
            Coords coords = new Coords(context);
            if (maxPt < 0) LH.errors.Add(new SemanticAnalysisError(coords,"No maxPt found!"));
            if (minSteps < 0) LH.errors.Add(new SemanticAnalysisError(coords,"No minsteps found!"));
            if (maxSteps < 0) LH.errors.Add(new SemanticAnalysisError(coords,"No maxsteps found!"));
            singleConfig = new SingleConfig(maxPt, minSteps, maxSteps);
        }
        public void EnterMaxpt([NotNull] TaskParser.MaxptContext context) { }
        public void ExitMaxpt([NotNull] TaskParser.MaxptContext context) { maxPt = int.Parse(context.NAT().GetText()); }
        public void EnterMinsteps([NotNull] TaskParser.MinstepsContext context) { }
        public void ExitMinsteps([NotNull] TaskParser.MinstepsContext context) { minSteps = int.Parse(context.NAT().GetText()); }
        public void EnterMaxsteps([NotNull] TaskParser.MaxstepsContext context) { }
        public void ExitMaxsteps([NotNull] TaskParser.MaxstepsContext context) { maxSteps = int.Parse(context.NAT().GetText()); }

        // Equation Rules
        public void EnterEq([NotNull] TaskParser.EqContext context) => LH.EnterEq();
        public void ExitEq([NotNull] TaskParser.EqContext context) => LH.ExitEq(new Coords(context));
        public void EnterEq_var_def([NotNull] TaskParser.Eq_var_defContext context) => LH.EnterVars();
        public void ExitEq_var_def([NotNull] TaskParser.Eq_var_defContext context) => LH.ExitEqVarDef();
        public void EnterFixed([NotNull] TaskParser.FixedContext context) => LH.EnterVars();
        public void ExitFixed([NotNull] TaskParser.FixedContext context) => LH.fixedVars = LH.adt_vars; 
        public void EnterForall([NotNull] TaskParser.ForallContext context) => LH.EnterVars(); 
        public void ExitForall([NotNull] TaskParser.ForallContext context) => LH.allQVars = LH.adt_vars;
        public void EnterVar([NotNull] TaskParser.VarContext context) { }
        public void ExitVar([NotNull] TaskParser.VarContext context)
            => LH.ExitVar(context.BEZ().GetText(), context.sort().BEZ().GetText(),new Coords(context));
        public void EnterSort([NotNull] TaskParser.SortContext context) { }
        public void ExitSort([NotNull] TaskParser.SortContext context) { }

        // Trees
        public void EnterMytree([NotNull] TaskParser.MytreeContext context) { }
        public void ExitMytree([NotNull] TaskParser.MytreeContext context) { }

        public void EnterEq_s([NotNull] TaskParser.Eq_sContext context) { }
        public void ExitEq_s([NotNull] TaskParser.Eq_sContext context) { }
        // classic Trees
        public void EnterClassicTree([NotNull] TaskParser.ClassicTreeContext context) { }
        public void ExitClassicTree([NotNull] TaskParser.ClassicTreeContext context) { }

        public void EnterTreeLeaf([NotNull] TaskParser.TreeLeafContext context) { }
        public void ExitTreeLeaf([NotNull] TaskParser.TreeLeafContext context)
            => LH.ExitTreeLeaf(context.BEZ().GetText(), new Coords(context));

        public void EnterTreeNode([NotNull] TaskParser.TreeNodeContext context) { }
        public void ExitTreeNode([NotNull] TaskParser.TreeNodeContext context)
            => LH.ExitTreeNode(context.BEZ().GetText(), context.ChildCount, context.GetText(),new Coords(context));
        // function Trees

        public void EnterFTree([NotNull] TaskParser.FTreeContext context) { }
        public void ExitFTree([NotNull] TaskParser.FTreeContext context) { }

        public void EnterFTreeRoot([NotNull] TaskParser.FTreeRootContext context) { }
        public void ExitFTreeRoot([NotNull] TaskParser.FTreeRootContext context)
        => LH.ExitFTreeRoot(new Coords(context));

        public void EnterFTreeLeft([NotNull] TaskParser.FTreeLeftContext context) { }
        public void ExitFTreeLeft([NotNull] TaskParser.FTreeLeftContext context) { }
        public void EnterFTreeRight([NotNull] TaskParser.FTreeRightContext context) { }
        public void ExitFTreeRight([NotNull] TaskParser.FTreeRightContext context) { }

        public void EnterFTreeBracket([NotNull] TaskParser.FTreeBracketContext context) { }
        public void ExitFTreeBracket([NotNull] TaskParser.FTreeBracketContext context) { }

        public void EnterFTreeBez([NotNull] TaskParser.FTreeBezContext context) { }

        public void ExitFTreeBez([NotNull] TaskParser.FTreeBezContext context)
            => LH.ExitFTreeBez(context.BEZ().GetText(), new Coords(context));

        // useless Interface requirements
        public void EnterEveryRule(ParserRuleContext ctx) { }
        public void ExitEveryRule(ParserRuleContext ctx) { }
        public void VisitErrorNode(IErrorNode node) { }
        public void VisitTerminal(ITerminalNode node) { }

        // Alle Offiziell notwendige Typtests

        private void Test_Lemmata_all_Unique(string lemmaName,Coords coords)
        {
            foreach (Rule axiom in LH.adt_axioms)
                if (axiom.name == lemmaName) LH.errors.Add(new SemanticAnalysisError(coords,"The Lemma " + lemmaName + " has same name like 1 Axiom of ADT!"));
            foreach (Rule lemma in lemmata)
                if (lemma.name == lemmaName) LH.errors.Add(new SemanticAnalysisError(coords,"The Lemma " + lemmaName + " is defined multiple times in Task!"));
            foreach (Rule lemma in lemmaTasks)
                if (lemma.name == lemmaName) LH.errors.Add(new SemanticAnalysisError(coords,"The Lemma " + lemmaName + " is defined multiple times in Task!"));
        }
    }
}
